/*-------------------------------------------------------------------*/
/* FastMST.c       ZHONG Caiming                                     */
/*                 23-08-2009                                        */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "FastMST"
#define VersionNumber  "Version 0.1"
#define LastUpdated    "04.09.2009"

/*-------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "FastMST.h"
#include "MSTKruskal.h"

#include "random.h"
#include "cb.h"
#include "memctrl.h"

/*-------------------------------------------------------------------*/
/* 
 * Randomly generate k centers, vectors of centers in BOOK are returned 
 */

float     * InitialCenters(TRAININGSET *TS, int bookSize, int k){

  float   * Centers = (float *)allocate(sizeof(float) * k * VectorSize(TS));
  int     * Queue = (int *)allocate(sizeof(int) * bookSize);
  int     i, j, num = 0;

  for (i = 0; i < bookSize; i ++){
    Queue[i] = 0;
  }

  while (num < k){
    for (i = 0; i < k - num; i ++){
      Queue[IRZ(bookSize)] = 1;
    }

    num = 0;
    for (i = 0; i < bookSize; i ++){
      if (Queue[i] == 1) num ++;
    }
  }

  num = 0;
  for (i = 0; i < bookSize; i ++){
    if (Queue[i] == 1){
      for (j = 0; j < VectorSize(TS); j ++){
        Centers[num * VectorSize(TS) + j] = TS->Book[i].vector[j];
      }
      num ++;
    }
  }
  
  deallocate(Queue);
  return Centers;
}

/* -------------------------------------------------------------------*/
/*
 *  Compute the distances from every point to changed centers
 */

void DistFromPointToCenter(TRAININGSET  * TS, 
                           float        * DistPC, 
                           float        * Centers, 
                           int          * Changed, 
                           int            NumChanged, 
                           int            k){

  int     bookSize  = BookSize(TS);
  int     vectSize  = VectorSize(TS);
  float * VectorTmp = (float *)allocate(sizeof(float) * vectSize);

  int     i, j, q;

  for (i = 0; i < bookSize; i ++){
    for (j = 0; j < NumChanged; j ++){  
      for (q = 0; q < VectorSize(TS); q ++){
        VectorTmp[q] = Centers[Changed[j] * vectSize + q];//corresponding center vector
      }

      EuclideanDistIF(TS->Book[i].vector, 
                      VectorTmp, 
                      vectSize, 
                      DistPC + i * k + Changed[j]);
    }
  }

  deallocate(VectorTmp);
  //return DistPC;
}

/* -------------------------------------------------------------------*/
/*
 *  Get new centroids of changed clusters
 */

void GetCentroids(TRAININGSET * TS, 
                  float       * Centers, 
                  int         * Partition, 
                  int         * Changed, 
                  int           NumChanged, 
                  int         * Counts){

  int     bookSize   = BookSize(TS);
  int     vectorSize = VectorSize(TS), count = 0;
  float * sum        = (float *)allocate(sizeof(float) * vectorSize);
  int     i, j, q;

  for (i = 0; i < NumChanged; i ++){
    count = 0;
    for (q = 0; q < vectorSize; q++){
      sum[q] = 0;
    }

    for (j = 0; j < bookSize; j ++){
      if(Partition[j] == Changed[i]){
        count++;
        for (q = 0; q < vectorSize; q++){
          sum[q] += TS->Book[j].vector[q]; 
        }
      }
    }

    Counts[Changed[i]] = count;
    for (q = 0; q < vectorSize; q++){    
      Centers[Changed[i] * vectorSize + q] = sum[q] / count;
    }
  }

  deallocate(sum);
}

/* -------------------------------------------------------------------*/
/*
 *  This KMeans algorithm is converted from Matlab 7.1.0.246 
 *
 *  The return is the pointer to cluster centers, while int * Partition
 *  is the partition result.
 *
 */

float   * KMeans_Matlab(TRAININGSET  * TS, 
                        int            k, 
                        int          * Partition, 
                        int            MaxIt){

  int     bookSize  = BookSize(TS);
  int     vectorSize = VectorSize(TS);

  int   * PrePartition = (int *)allocate(sizeof(int) * bookSize);
  int   * CurrentPartition = (int *)allocate(sizeof(int) * bookSize); 
  int   * TmpPartition = (int *)allocate(sizeof(int) * bookSize);
  float * CurrentMin = (float*)allocate(sizeof(float) * bookSize);
  float * MinDel = (float*)allocate(sizeof(float) * bookSize);
  int   * Counts = (int *)allocate(sizeof(int) * k); // m

  float * DistPC = (float*)allocate(sizeof(float) * bookSize * k);
  float * Del = (float*)allocate(sizeof(float) * bookSize * k);
  float * Centers;//C

  int   * Changed = (int *)allocate(sizeof(int) * k);
  int   * Moved = (int *)allocate(sizeof(int) * bookSize);
  int   * Mbrs = (int *)allocate(sizeof(int) * bookSize);
  int   * Sgn = (int *)allocate(sizeof(int) * bookSize);

  int     NumChanged = k, NumMoved = k;
  int     LastMoved, NumLastMoved; 
  int     i, j, q, converged = 0;

  int     iter = 0, iter1;
  float   totalSumD = -1, preTotalSumD = -1;
  int     oidx, nidx;
  float   tmp, tmp1;

  //initialization
  for (i = 0; i < k; i ++){
    Changed[i] = i;
  }

  for (i = 0; i < bookSize; i ++){
    Partition[i] = 0;
    PrePartition[i] = 0;
    CurrentPartition[i] = 0;
    TmpPartition[i] = 0;
  }

  Centers = InitialCenters(TS, bookSize, k);

  for(;;)
  {
    //Compute the distance from every point to changed cluster centroid
    DistFromPointToCenter(TS, DistPC, Centers, Changed, NumChanged, k);

    //Compute the total sum of distances for the current configuration
    //Can't do it first time through, there's no configuration yet
    if (iter > 0){
      totalSumD = 0;
      for (i = 0; i < bookSize; i ++){
        totalSumD += DistPC[i * k + Partition[i]];
      }

      //Test for a cycle: if objective is not decreased, back out
      //the last step and move on to the single update phase
      if (preTotalSumD <= totalSumD && preTotalSumD != -1){
        for (i = 0; i < bookSize; i ++){
          Partition[i] = PrePartition[i];
        }

        GetCentroids(TS, Centers, Partition, Changed, NumChanged, Counts);
        iter = iter - 1;
        break;        
      }

      if (iter >= MaxIt) break;
    }

    //Determine closest cluster for each point and reassign points to clusters
    for (i = 0; i < bookSize; i ++){
      PrePartition[i] = Partition[i];
    }

    preTotalSumD = totalSumD;
    for (i = 0; i < bookSize; i ++){
      tmp = -1;
      for (j = 0; j < k; j ++){
        if (tmp > DistPC[i * k + j] || tmp < 0){
          CurrentPartition[i] = j;
          tmp = DistPC[i * k + j];
        }
      }
      CurrentMin[i] = tmp;
    }

    if (iter == 0){
      //Every point moved, every cluster will need an update
      for (i = 0; i < bookSize; i ++){
        Moved[i] = i;
        Partition[i] = CurrentPartition[i];
      }      
      
      for (i = 0; i < k; i ++){
        Changed[i] = i;
      }
    }
    else
    {
      //Determine which points moved
      NumMoved = 0;
      for (i = 0; i < bookSize; i ++){
        if (CurrentPartition[i] != PrePartition[i]){
          Moved[NumMoved] = i;
          NumMoved ++;
        }
      }

      j = 0;
      if (NumMoved > 0){
        for (i = 0; i < NumMoved; i ++){
          if (DistPC[Moved[i] * k + PrePartition[Moved[i]]] > CurrentMin[Moved[i]]){
            Moved[j++] = Moved[i];
          }
        }
      }

      NumMoved = j;
      if (j == 0) break;

      for (i = 0; i < NumMoved; i ++){
        Partition[Moved[i]] = CurrentPartition[Moved[i]];
      }

      //Find clusters that gained or lost members
      for (i = 0; i < bookSize; i ++){
        TmpPartition[i] = 0;
      }

      for (i = 0; i < NumMoved; i ++){
        if (TmpPartition[Partition[Moved[i]]] == 0){
          TmpPartition[Partition[Moved[i]]] = 1;
        }
        if (TmpPartition[PrePartition[Moved[i]]] == 0){
          TmpPartition[PrePartition[Moved[i]]] = 1;
        }
      }

      NumChanged = 0;
      for (i = 0; i < bookSize; i ++){
        if (TmpPartition[i] == 1){
          Changed[NumChanged++] = i;
        }
      }
    }


    //Calculate the new cluster centroids and counts
    GetCentroids(TS, Centers, Partition, Changed, NumChanged, Counts);
    iter ++;

    //Deal with clusters that have just lost all their members
    //Skipped
  }

  //Begin phase two:  single reassignments
  NumChanged = 0;
  for (i = 0; i < k; i ++){
    if (Counts[i] > 0){
      Changed[NumChanged++] = i;
    }
  }

  LastMoved = 0;
  NumLastMoved = 0;
  iter1 = iter;

  while (iter < MaxIt)
  {
    for (i = 0; i < NumChanged; i ++){
      for (j = 0; j < bookSize; j ++){
        Mbrs[j] = (Partition[j] == Changed[i]);
        Sgn[j]  = 1- 2 * Mbrs[j];
      }
      
      if (Counts[Changed[i]] == 1){
        for (j = 0; j < bookSize; j ++){
          Sgn[Mbrs[j]] = 0;
        }
      }

      for (j = 0; j < bookSize; j ++){
        tmp1 = 0;
        for (q = 0; q < vectorSize; q ++){
          tmp   = TS->Book[j].vector[q] - Centers[Changed[i] * vectorSize + q];
          tmp1 += tmp * tmp;
        }
        Del[j * k + Changed[i]] = ((float)(Counts[Changed[i]])) / (Counts[Changed[i]] + Sgn[j]) * tmp1;
      }
    }

    //Determine best possible move, if any, for each point.  Next we
    //will pick one from those that actually did move.
    for (i = 0; i < bookSize; i ++){
      PrePartition[i] = Partition[i];
    }  

    preTotalSumD = totalSumD;

    for (i = 0; i < bookSize; i ++){
      tmp = -1;
      for (j = 0; j < k; j ++){
        if (tmp > Del[i * k + j] || tmp < 0){
          CurrentPartition[i] = j;
          tmp = Del[i * k + j];
        }
      }
      MinDel[i] = tmp;
    }

    //Determine which points moved
    NumMoved = 0;
    for (i = 0; i < bookSize; i ++){
      if (PrePartition[i] != CurrentPartition[i]){
        Moved[NumMoved] = i;
        NumMoved ++;
      }
    }

    j = 0;
    if (NumMoved > 0){
      for (i = 0; i < NumMoved; i ++){
        if (Del[Moved[i] * k + PrePartition[Moved[i]]] > MinDel[Moved[i]]){
          Moved[j++] = Moved[i];
        }
      }
    }

    NumMoved = j;
    if (NumMoved == 0){
      if ((iter - iter1) == 0 || NumLastMoved > 0){
        iter = iter + 1;
      }
      converged = 1;
      break;
    }

    //Pick the next move in cyclic order
    tmp = -1;
    for (i = 0; i < NumMoved; i ++){
      tmp1 = (Moved[i] - LastMoved - 1) % bookSize + LastMoved;
      if (tmp > tmp1 || tmp == -1){
        tmp = tmp1;
      }
    }

    tmp = (int)tmp % bookSize;
    if (tmp < 0) tmp = 0;

    Moved[0] = tmp;
    NumMoved = 1;

    if (Moved[0] <= LastMoved){
        iter = iter + 1;
        NumLastMoved = 0;
        if (iter > MaxIt){
          break;
        }
    }

    NumLastMoved = NumLastMoved + 1;
    LastMoved    = Moved[0];
    
    oidx         = Partition[Moved[0]];
    nidx         = CurrentPartition[Moved[0]];
    totalSumD    = totalSumD + Del[Moved[0] * k + nidx] - Del[Moved[0] * k + oidx];

    // Update the cluster index vector, and the old and new cluster
    // counts and centroids

    Partition[Moved[0]] = nidx;
    Counts[nidx] = Counts[nidx] + 1;
    Counts[oidx] = Counts[oidx] - 1;

    for (i = 0; i < vectorSize; i ++){
      Centers[nidx * vectorSize + i] = Centers[nidx * vectorSize + i] +
                                       (TS->Book[Moved[0]].vector[i] - Centers[nidx * vectorSize + i]) / Counts[nidx];
    
      Centers[oidx * vectorSize + i] = Centers[oidx * vectorSize + i] -
                                       (TS->Book[Moved[0]].vector[i] - Centers[oidx * vectorSize + i]) / Counts[oidx];
    }

    if (oidx > nidx){
      Changed[0] = nidx;
      Changed[1] = oidx;
    }
    else{
      Changed[0] = oidx;
      Changed[1] = nidx;    
    }
    NumChanged = 2;
  }//end phase two

  deallocate(PrePartition);
  deallocate(CurrentPartition);
  deallocate(TmpPartition);
  deallocate(CurrentMin);
  deallocate(MinDel);
  deallocate(Counts);
  deallocate(Changed);
  deallocate(Moved);
  deallocate(Mbrs);
  deallocate(Sgn);
  deallocate(DistPC);
  deallocate(Del);

  return Centers;
}

/*-------------------------------------------------------------------*/

int   * FastMSTOneRound(TRAININGSET * TS, 
                        int           k, 
                        float       * Centroids, 
                        int         * Partition,
                        float       * CentersOfCentroids){

  int             vectorSize = VectorSize(TS), bookSize = BookSize(TS);
  TRAININGSET  ** SubTS;
  int          ** PPairResult = (int **) allocate(sizeof(int *) * k); 
  int          ** PPairResultTransformed = (int **) allocate(sizeof(int *) * k);  
  int          ** PMappings = (int **) allocate(sizeof(int *) * k); 
  float        ** CentroidVectorSet;
  int           * PairResultCentroid;
  int           * PFastMST;
  int           * SubTSSize = (int *)allocate(sizeof(int) * k);

  int           * PairResultFastMST = (int *) allocate(sizeof(int) * (bookSize - 1) * 3 );

  float         * VectorFloatTmp, dist, minDist;
  int             i, j, q, idx1, idx2, nearestIdx1, nearestIdx2;

  //1. Construct the mapping between subset indices and whole data set indices.

  //1.1.1 Size of each subset
  for (i = 0; i < k; i ++){
    SubTSSize[i] = 0;
  }

  for (i = 0; i < bookSize; i ++){
    SubTSSize[Partition[i]]++;
  }

  //1.1.2 Allocate memory for Mapping
  for (i = 0; i < k; i ++){
    PMappings[i] = (int *)allocate(sizeof(int) * SubTSSize[i]);
  }

  //1.2 Mapping
  for (i = 0; i < bookSize; i ++){
    *(PMappings[Partition[i]]) = i;
    (PMappings[Partition[i]])++; 
  }

  //1.2.1 Move back the pointer of Mapping
  for (i = 0; i < k; i ++){
    PMappings[i] -= SubTSSize[i];
  }
  
  //2. MST of each subset
  //2.1 Construct sub Training sets
  SubTS = (TRAININGSET **)allocate(sizeof(TRAININGSET *) * k);

  //2.1.1 Allocate memory for k subsets 
  for (i = 0; i < k; i ++){
    SubTS[i] = (TRAININGSET *)allocate(sizeof(TRAININGSET));
    SubTS[i]->BlockSizeX = TS->BlockSizeX;
    SubTS[i]->BlockSizeY = TS->BlockSizeY;
    SubTS[i]->CodebookSize = SubTSSize[i];
    SubTS[i]->Book = (BOOKNODE *)allocate(sizeof(BOOKNODE) * SubTSSize[i]);
  }

  //2.1.2 Write subsets
  for (i = 0; i < bookSize; i ++){    
    *(SubTS[Partition[i]]->Book++) = TS->Book[i];
  }

  //2.1.3 Move back the pointer of Book
  for (i = 0; i < k; i ++){
    SubTS[i]->Book -= SubTSSize[i];
  }

  //2.2 MST on k subsets
  for (i = 0; i < k; i ++){
    PPairResult[i] = MST_KruskalFromTrainingSetReturnPair(SubTS[i]);
  }

  //2.3 Convert the local indices of subset MST into global indices
  for (i = 0; i < k; i ++){
    PPairResultTransformed[i] = (int *)allocate(sizeof(int) * (SubTSSize[i] - 1) * 3);
  }

  for (i = 0; i < k; i ++)
    for (j = 0; j < (SubTSSize[i] - 1) * 3; j += 3){
      PPairResultTransformed[i][j] = PMappings[i][PPairResult[i][j]];
      PPairResultTransformed[i][j + 1] = PMappings[i][PPairResult[i][j + 1]];
    }


  //2.4 Combine the pairs of subset MSTs
  PFastMST = PairResultFastMST;
  for (i = 0; i < k; i ++){
    for (j = 0; j < SubTSSize[i] - 1; j ++){
      *PFastMST++ = PPairResultTransformed[i][j * 3];
      *PFastMST++ = PPairResultTransformed[i][j * 3 + 1];
      PFastMST++;
    }
  }  

  //3 MST on centriods
  //3.1 Construct CentriodVectorSize
  CentroidVectorSet = (float **)allocate(sizeof(float *) * k);

  for (i = 0; i < k; i ++){    
    VectorFloatTmp = (float *)allocate(sizeof(float) * vectorSize);
    for (j = 0; j < vectorSize; j ++){
      VectorFloatTmp[j] = Centroids[i * vectorSize + j];
    }

    CentroidVectorSet[i] = VectorFloatTmp;
  }

  //3.2 MST of centroids
  PairResultCentroid = MST_KruskalFromVectorSetReturnPair(CentroidVectorSet, 
                                                          vectorSize, 
                                                          k);

  //3.3 Determine the edges between subsets and combine,
  //and produce CentersOfCentroids
  VectorFloatTmp = (float *)allocate(sizeof(float) * vectorSize);
  for (i = 0; i < k - 1; i ++){
    idx1 = PairResultCentroid[i * 3];
    idx2 = PairResultCentroid[i * 3 + 1];

    //produce CentersOfCentroids
    for (j = 0; j < vectorSize; j ++){
      CentersOfCentroids[i * vectorSize + j] = (Centroids[idx1 * vectorSize + j] +
                                                Centroids[idx2 * vectorSize + j]) / 2;
    }

    //Find the nearest pairs
    minDist = -1;  
    if (SubTSSize[idx1] > 1){
      for (j = 0; j < SubTSSize[idx1]; j ++){
        for (q = 0; q < vectorSize; q ++){
          VectorFloatTmp[q] = Centroids[idx2 * vectorSize + q];
        }

        EuclideanDistIF(TS->Book[PMappings[idx1][j]].vector, 
                        VectorFloatTmp, 
                        VectorSize(TS),
                        &dist);

        if (minDist == -1 || dist < minDist){
          minDist = dist;
          nearestIdx1 = PMappings[idx1][j];
        }
      }
    }
    else{
      nearestIdx1 = PMappings[idx1][0];
    }

    minDist = -1;
    if (SubTSSize[idx2] > 1){
      for (j = 0; j < SubTSSize[idx2]; j ++){
        for (q = 0; q < vectorSize; q ++)
          VectorFloatTmp[q] = Centroids[idx1 * vectorSize + q];

        EuclideanDistIF(TS->Book[PMappings[idx2][j]].vector, 
                        VectorFloatTmp, 
                        VectorSize(TS),
                        &dist);

        if (minDist == -1 || dist < minDist){
          minDist = dist;
          nearestIdx2 = PMappings[idx2][j];
        }
      }
    }
    else{
      nearestIdx2 = PMappings[idx2][0];
    }

    //Combine the edges
    *PFastMST++ = nearestIdx1;
    *PFastMST++ = nearestIdx2;
    PFastMST++;
  }


  deallocate(VectorFloatTmp);
                           
  for (i = 0; i < k; i++){
    deallocate(PMappings[i]);
    deallocate(CentroidVectorSet[i]);
    deallocate(SubTS[i]->Book);
    deallocate(SubTS[i]);
    deallocate(PPairResultTransformed[i]);
    deallocate(PPairResult[i]);
  }

  deallocate(CentroidVectorSet);  
  deallocate(SubTSSize);
  deallocate(SubTS);
  deallocate(PPairResultTransformed);
  deallocate(PMappings);
  deallocate(PPairResult);
  deallocate(PairResultCentroid);

  return PairResultFastMST;
}

/*-------------------------------------------------------------------*/
/*
 * This function aims at partitioning points in near boundaries into
 * same clusters.
 */

int * BoundaryBasedPartition(TRAININGSET  * TS, 
                             int            k, 
                             float        * CentersOfCentroids){

  int       bookSize = BookSize(TS);
  int       vectorSize = VectorSize(TS);
  int     * Partition = (int *)allocate(sizeof(int) * bookSize);
  float   * VectorTmp = (float *)allocate(sizeof(float) * vectorSize);
  int       i, j, q, idx;
  float     dist, minDist;
  
  for (i = 0; i < bookSize; i ++){
    minDist = -1;
    for (j = 0; j < k - 1; j ++){//k - 1 partitions
      for (q = 0; q < vectorSize; q ++){
        VectorTmp[q] = CentersOfCentroids[j * vectorSize + q];
      }    

      EuclideanDistIF(TS->Book[i].vector, VectorTmp, vectorSize, &dist);
      if (minDist > dist || minDist < 0){
        minDist = dist;
        idx = j;
      }
    }
    Partition[i] = idx;
  }

  deallocate(VectorTmp);
  return Partition;
}

/*-------------------------------------------------------------------*/

int RemoveEmptyPartition(float  * Centroids, 
                         int    * Partition, 
                         int      k, 
                         int      bookSize, 
                         int      vectorSize){
  int    len = 0;
  int    i, j;
  int  * Tag = (int *)allocate(sizeof(int) * k);

  for (i = 0; i < k; i ++){
    Tag[i] = 0;
  }

  for (i = 0; i < bookSize; i ++){
    Tag[Partition[i]] = 1;
  }

  for (i = 0; i < k; i ++){
    if (Tag[i] == 0){
      for (j = 0; j < bookSize; j ++){
        if (Partition[j] > i - len)Partition[j]--;
      }

      for (j = i - len + 1; j < k - len; j ++){
        Centroids[(j - 1) * vectorSize] =Centroids[j * vectorSize];
        Centroids[(j - 1) * vectorSize + 1] =Centroids[j * vectorSize + 1];
      }  
      len ++;
    }
  }

  deallocate(Tag);
  return len;
}

/*-------------------------------------------------------------------*/
/*  k: number of partitions, num: number of rounds of FastMST
 *  The return pointer is MST Pairs, and can be easily converted
 *  to GRAPH
 */

int * FastMST(TRAININGSET *TS, int k, int numOfRounds){

  int         bookSize = BookSize(TS);
  int         vectorSize = VectorSize(TS);
  char      * Matrix = (char *)allocate(sizeof(char) * bookSize * bookSize);// to be improved!
  Pair      * Pairs = (Pair*)allocate(sizeof(Pair) * bookSize * numOfRounds * 2);//two rounds: Kmeans + boundary
  PairInfo  * PInfo = (PairInfo*)allocate(sizeof(PairInfo));
  int       * PairResultFastMST1;
  int       * PairResultFastMST2;
  int       * PairResultFinal;
  int       * CenterPartition;
  int       * Partition = (int *)allocate(sizeof(int) * bookSize);
  float     * Centroids;
  float     * CentersOfCentroids;
  
  FILE      * fp;
  
  int i, j, idx1, idx2, numPairs = 0, numEmptyPartition1 = 0, numEmptyPartition2 = 0;

  //num rounds of KMeans-MST
  for (i = 0; i < numOfRounds; i ++){
    //KMeans partition
    if (i > 0) {
      deallocate(Centroids);
      deallocate(PairResultFastMST1);
      deallocate(PairResultFastMST2);
      deallocate(CentersOfCentroids);
      deallocate(CenterPartition);
    }

    Centroids          = KMeans_Matlab(TS, k, Partition, MAXIT);
    numEmptyPartition1 = RemoveEmptyPartition(Centroids, 
                                              Partition, 
                                              k, 
                                              bookSize, 
                                              vectorSize);
    CentersOfCentroids = (float *) allocate(sizeof(float) * (k - numEmptyPartition1 - 1) * vectorSize);
    PairResultFastMST1 = FastMSTOneRound(TS, 
                                         k - numEmptyPartition1, 
                                         Centroids, 
                                         Partition,
                                         CentersOfCentroids);

    //Boundary-oriented partition
    CenterPartition    = BoundaryBasedPartition(TS, k - numEmptyPartition1, CentersOfCentroids);
    numEmptyPartition2 = RemoveEmptyPartition(CentersOfCentroids, 
                                              CenterPartition, 
                                              k - numEmptyPartition1 - 1, 
                                              bookSize, 
                                              vectorSize);

    PairResultFastMST2 = FastMSTOneRound(TS, 
                                         k - numEmptyPartition1 - numEmptyPartition2 - 1, 
                                         CentersOfCentroids, 
                                         CenterPartition,
                                         CentersOfCentroids);

    //Combine the result
    for (j = 0; j < bookSize - 1; j ++){
      idx1 = PairResultFastMST1[j * 3];
      idx2 = PairResultFastMST1[j * 3 + 1];

      if (Matrix[idx1 * bookSize + idx2] != 1){
        Pairs[numPairs].index1 = idx1;
        Pairs[numPairs].index2 = idx2;
        EuclideanDistII(TS->Book[idx1].vector, 
                        TS->Book[idx2].vector,                                         
                        vectorSize, 
                        &(Pairs[numPairs].weight));

        Matrix[idx1 * bookSize + idx2] = 1;
        Matrix[idx2 * bookSize + idx1] = 1;
        numPairs++;
      }
    }

    for (j = 0; j < bookSize - 1; j ++){
      idx1 = PairResultFastMST2[j * 3];
      idx2 = PairResultFastMST2[j * 3 + 1];

      if (Matrix[idx1 * bookSize + idx2] != 1){
        Pairs[numPairs].index1 = idx1;
        Pairs[numPairs].index2 = idx2;
        EuclideanDistII(TS->Book[idx1].vector, 
                        TS->Book[idx2].vector,                                         
                        vectorSize, 
                        &(Pairs[numPairs].weight));

        Matrix[idx1 * bookSize + idx2] = 1;
        Matrix[idx2 * bookSize + idx1] = 1;
        numPairs++;
      }
    }  
  }

  //PInfo
  PInfo->Pairs    = Pairs;
  PInfo->PairSize = numPairs;
  PairResultFinal = MST_KruskalFromPairsReturnPair(TS, PInfo);

  deallocate(Centroids);
  deallocate(Partition);
  deallocate(Matrix);
  deallocate(PInfo);
  deallocate(Pairs);
  deallocate(PairResultFastMST1);
  deallocate(PairResultFastMST2);
  deallocate(CenterPartition);
  deallocate(CentersOfCentroids);

  return PairResultFinal;
}

