/*-------------------------------------------------------------------*/
/* FASTKNN.C           Mikko Malinen                                 */
/*                                                                   */
/*                                                                   */
/* Calculates FastkNN graph. First divides the trainingset into      */
/* clusters and then makes kNN graph for each cluster. At last the   */
/* kNN graphs are combined to final kNN.                             */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "FASTKNN"
#define VersionNumber  "Version 0.03"
#define LastUpdated    "13.02.17" /* Sami Sieranoja */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* MM     0.01  Initial version from 25.8.09                         */
/* SS     0.03  Compile with new modules dir 13.02.17                */
/*                                                                   */
/* ----------------------------------------------------------------- */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>



#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "owntypes.h"
#include "interfc.h"
#include "random.h"
#include "reporting.h"
#include "rs.h"
#include "pnngraph.h"
#include "kmeans.h"
#include "FastkNN.h"


/* ------------------------------------------------------------------ */

/* from kmeans.c */
static void GenerateSolution(TRAININGSET *pTS, CODEBOOK *pCBnew,
                             PARTITIONING *pPnew, CODEBOOK *pCBinit,
                             PARTITIONING *pPinit, llong *distance,
                             llong *distanceInit, int InitMethod, int initial)
{
  int i;

  if (initial)
    {
      CopyCodebook(pCBinit, pCBnew);
      CopyPartitioning(pPinit, pPnew);
      for (i = 0; i < BookSize(pTS); i++)
	{
	  distance[i] = distanceInit[i];
	}
    }
  else
    {
      if (InitMethod == 0)
	SelectRandomRepresentatives(pTS, pCBnew);
      else if (InitMethod == 1)
	SelectRandomWeightedRepresentatives(pTS, pCBnew);
      else if (InitMethod == 2)
	SelectRandomRepresentativesbyMarko(pTS, pCBnew);
      else
	SelectRandomRepresentatives2(pTS, pCBnew);

      GenerateOptimalPartitioningGeneral(pTS, pCBnew, pPnew, MSE);
      CalculateDistances(pTS, pCBnew, pPnew, distance);
    }
}  /* GenerateSolution() */



Graph * FastkNN(TRAININGSET*       TS,
		  CODEBOOK*        CB,
		  PARTITIONING*    P,
		  int              K,
		  int              FinalSize,
		  int              Heap,
		  int              GCMethod,
		  int              Quiet,
		  long             watch,
		  int              Pds,
		  int              Mps,
		  int              Mlf,
		  int              Lazy,
		  int              Cf)
{
  int           numberOfClustersInFast;
  const int     KmeansIterationsInFast = 10; /* change if necessary */
  PARTITIONING  Pnew, Pinit;
  CODEBOOK      CBnew, CBinit;
  double        error;
  CODEBOOK      CBFast;
  PARTITIONING  PFast;
  llong         distance[BookSize(CB)];
  llong         distanceInit[BookSize(CB)];
  int           i, iter;
  int           active[(int)(sqrt(BookSize(CB))+0.5)];
  int           activeCount;
  CODEBOOK      CBsmall;
  PARTITIONING  Psmall;
  Graph*        g;
  Graph*        g2;
  GraphVector*  v;
  int           j, k, m;
  int           neighborMap[BookSize(CB)];
  llong         cdist[BookSize(CB)];

  numberOfClustersInFast = (int)(sqrt(BookSize(CB))+0.5);
  error = 0;
  if ((numberOfClustersInFast < 1) || (BookSize(CB) < numberOfClustersInFast))
    {
      return 0;   /* clustering failed */
    }
  CreateNewCodebook(&CBFast, numberOfClustersInFast, CB);
  CreateNewPartitioning(&PFast, CB, numberOfClustersInFast);

  /* InitializeSolutions */
  CreateNewCodebook(&CBnew, numberOfClustersInFast, CB);
  CreateNewPartitioning(&Pnew, CB, numberOfClustersInFast);
  CreateNewCodebook(&CBinit, numberOfClustersInFast, CB);
  CreateNewPartitioning(&Pinit, CB, numberOfClustersInFast);


  iter = 0;
  GenerateSolution(CB, &CBnew, &Pnew, &CBinit, &Pinit, distance,
		   distanceInit, 0, 0);

  for (i = 0;i<KmeansIterationsInFast;i++)
    {
      OptimalRepresentatives(&Pnew, CB, &CBnew, active, cdist, &activeCount,NO);
      OptimalPartition(&CBnew, CB, &Pnew, active, cdist, activeCount, distance,NO);
    }

  error = AverageErrorForSolution(CB, &CBnew, &Pnew, MSE);

  CopyCodebook(&CBnew, &CBFast);
  CopyPartitioning(&Pnew, &PFast);

  FreeCodebook(&CBnew);
  FreePartitioning(&Pnew);
  FreeCodebook(&CBinit);
  FreePartitioning(&Pinit);

  /* Create the big graph */
  g = AllocateMemoryForGraph(BookSize(CB), K, VectorSize(CB));

  for ( i = 0; i < BookSize(CB); i++ ) {
    v = AllocateMemoryForGraphVector(g, i);
    for ( j = 0; j < VectorSize(CB); j++) {
      GraphPutVectorElement(v, j, VectorScalar(CB,i,j));
    }
  g->vectors[i] = v;
  }

  /* check that there is enough vectors in each cluster for kNN */



  /* Find kNN graph for each cluster */
  for (i=0; i<numberOfClustersInFast; i++)
    {
      m = 0;
      for (j=0; j<BookSize(CB); j++)
        {
          if (Map(&PFast,j)==i)
            {
              neighborMap[m++] = j;
            }
        }

      CreateNewCodebook(&CBsmall, m, CB);
      k=0;
      for (j=0; j<BookSize(CB); j++)
	{
	  if (Map(&PFast,j)==i)
	    {
              CopyVector(Vector(CB, j), Vector(&CBsmall,k), VectorSize(CB));
	      VectorFreq(&CBsmall,k) = VectorFreq(CB,j);
	      k++;
	    }
	}
      CreateNewPartitioning(&Psmall, &CBsmall, BookSize(&CBsmall));
      PutAllInOwnPartition(&CBsmall, &Psmall);
      g2 = PNNGraph2(&CBsmall,
		   &CBsmall,
		   &Psmall,
		   K,
		   BookSize(&CBsmall),
		   Heap,
		   GCMethod,
		   Quiet,
		   watch,
		   Pds,
		   Mps,
		   Mlf,
		   Lazy,
		   Cf);

      for(j=0; j<m; j++)
	{
	  for (k=0; k<K; k++)
	    {
             g->vectors[neighborMap[j]]->kindices[k] =
	       neighborMap[g2->vectors[j]->kindices[k]];
	    }
	}
      FreeGraph(g2);
      FreeCodebook(&CBsmall);
      FreePartitioning(&Psmall);
    }

  return g;
  /*return(0);*/ /* if error */
}
