/*-------------------------------------------------------------------*/
/* MST_Prim.c                       Mikko Malinen                    */
/*                                                                   */
/* Minimum spanning tree calculation using Prim's algorithm.         */
/* Returns pointer to graph.                                         */
/* Based on Olli Virmajoki's routine 02.08.09                        */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "MST_Prim"
#define VersionNumber  "Version 0.01"
#define LastUpdated    "07.08.09"  /*MM*/

/*-------------------------------------------------------------------*/


#include <stdio.h>

#include "cb.h"
#include "pnngraph.h"
/*# include "pnnfast.h" */
#include "memctrl.h"
#include "MST_Prim.h"

#define  round(a)         ((a)+0.5)



/*--------------------FUNCTIONS---------------------------------------*/

/* modified from pnngraph.c CreateDataStructures() */
static void InitializeKNN(KNNDATA *KNN, CODEBOOK *CB, int K)
{
  int i,j;
  KNN->NN = (NEARESTTYPE**)allocate(BookSize(CB) * sizeof(NEARESTTYPE*));
  KNN->head = (VERTEX**)allocate(BookSize(CB)*sizeof(VERTEX*));
  for( i = 0; i < BookSize(CB); i++ )
    {
      KNN->NN[i] = (NEARESTTYPE*)allocate(K * sizeof(NEARESTTYPE));
      KNN->head[i] = NULL;
      for( j = 0; j < K; j++ )
	{
	  KNN->NN[i][j].VectorDist = MAXLLONG/10;
	  KNN->NN[i][j].Vector = i;
	  KNN->NN[i][j].Recalculate = NO;
	}
    }

}

/* modified from pnngraph.c PnnGraphWriteGraph() */
static Graph *FormGraph(CODEBOOK* CB, KNNDATA* KNN, int K,                                             int FinalSize)
{
  int i, j, currcb = 0, *lut;
  Graph *g_out;
  GraphVector *v;

  lut = (int *) allocate(sizeof(int)*BookSize(CB));

  g_out = AllocateMemoryForGraph(FinalSize, K, VectorSize(CB));


  for ( i = 0; i < BookSize(CB); i++) {
    if (VectorFreq(CB,i) != 0) {
      lut[i] = currcb;
      printf("%d %d\n", i, currcb);
      currcb++;
    } else {
      lut[i] = -1;
    }
  }

  for ( i = 0; i < BookSize(CB); i++ ) {
    if (VectorFreq(CB,i) != 0) {
      v = AllocateMemoryForGraphVector(g_out, lut[i]);
      for ( j = 0; j < VectorSize(CB); j++) {
        GraphPutVectorElement(v, j, VectorScalar(CB,i,j));
      }
      for ( j = 0; j < K; j++ ) {
        GraphPutNN(g_out, v, lut[KNN->NN[i][j].Vector], j);
      }
      printf("%p  %p\n", v, g_out->vectors[lut[i]]);
      g_out->vectors[lut[i]] = v;
    }
  }
  return g_out;
}

static void FreeKNN(KNNDATA* KNN, int origCBsize)
{
  int i;
  VERTEX* p;
  VERTEX* pe;
  for( i = 0; i < origCBsize; i++ )
    {
      deallocate(KNN->NN[i]);
    }
  deallocate(KNN->NN);
  for(i = 0; i < origCBsize; i++ )
    {
      p=KNN->head[i];
      while(p!=NULL)
	{
	  pe=p;
	  p=p->link;
	  deallocate(pe);
	}
    }
  deallocate(KNN->head);
}

/* from pnngraph.c */
static llong MergeDistortion(CODEBOOK* CB, int p1, int p2)
{
  /* Original "inline" */
  int i=0, Vsize;
  llong   dist=0LL, diff, dist_limit=MAXLLONG;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
    ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  /* Precondition: Vsize >= 1 */
  /* if dist >= dist_limit, returns at least factor * dist_limit */
  do
    {
      diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
      dist += diff * diff;
      /*       STATCODE
	 (
	  ksum++; / * Statistical information * /
	  )
      */
	 } while(dist < dist_limit && ++i < Vsize);
  dist=round( factor * (double)dist);
  /*  STATCODE
    (
     ndistcalcs++; / * Statistical information * /
     )
  */
    return( dist);
}



Graph*  MST_Prim(CODEBOOK* CB, /* codebook */ 
		     /*KNNDATA* KNN,*/ /* distance matrix */
		     /*KDATA* KD, */   /* kd-tree data */
		     /*HEAP* H,  */    /* heap */
		     int Quiet)    /* quietlevel */
{
   KNNDATA* KNN;
   Graph* g;
   int N = BookSize(CB);
   llong *lowcost;
   int *closest;
   int *connected;
   int i, j, k=0;
   llong min, Dist;
   
   lowcost = (llong*)allocate(N * sizeof(llong));
   closest = (int*)allocate(N * sizeof(int));
   connected = (int*)allocate(N * sizeof(int));
   
   /*Initializations:*/
   connected[0] = 1;
   closest[0] = 0;
   lowcost[0] = MAXLLONG;
   for(i=1; i<N; ++i)
   {
      connected[i] = 0;
      closest[i] = 0;
      lowcost[i] = MergeDistortion(CB, 0, i);
   }
   InitializeKNN(KNN, CB, 1);

   /* Identify the MST */
   for(i=1; i<N; ++i)
   {
      /* Find the k not in the tree closest to a vertex in the tree */
      min = MAXLLONG;
      for(j=1; j<N; ++j)
         if(!connected[j] && lowcost[j] < min)
         {
            min = lowcost[j];
            k = j;
         }
      /* Include k */
      connected[k]=1;

      /* Update lowcost[] and closest[] */
      for(j=1; j<N; ++j)
      {
		 Dist = MergeDistortion(CB,k,j);
         if( !connected[j] && Dist < lowcost[j] )
         {
			 lowcost[j] =  Dist;
			 closest[j] = k;         	
         }
      }
   }

   for(i=1;i<N;++i)
   {
		KNN->NN[i][0].VectorDist=lowcost[i];
		KNN->NN[i][0].Vector=closest[i];
		if(Quiet>2) 
			printf("\n%d -> %d dist %f",i,closest[i],(double)lowcost[i]);
		/*
		DM_distset(KD, i, KNN->NN[i][0].VectorDist);
		DM_toset(KD, i, closest[i]);
		Heap_insert(&(KD->H), DM_elem(KD, i), NULL, DM_heapindexaddr(KD, i));
		*/
   }

   g = FormGraph(CB, KNN, 1, BookSize(CB));

   FreeKNN(KNN, BookSize(CB));
   deallocate(lowcost);
   deallocate(closest);
   deallocate(connected);

   return g;
}


/*-------------------------------------------------------------------*/
