/*
$Revision $
$Date $
$Author $
$Name $
$Id $
*/

/*-------------------------------------------------------------------*/
/* NEIGHB.C        Marko Tuononen                                    */
/*                                                                   */
/* Finding neighbour clusters                                        */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "NEIGHB"
#define VersionNumber   "Version 0.11"  /* mm */
#define LastUpdated     "10.08.2009"

/* -------------------------CHANGE LOG------------------------------ */
/*                                                                   */
/*  MM 0.11  Graph forming routine added                             */
/*  MT 0.1   Previous version                                        */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include <float.h>

#include "cb.h"
#include "memctrl.h"
#include "interfc.h"
#include "stack.h"

/* ====================== CRUNCHING ROUTINES ========================= */

double* CalculateVectorAverage(VECTORTYPE v1, VECTORTYPE v2, int Vsize) 
{
  int i;
  double *vec; 
  vec = (double*)allocate(sizeof(double)*Vsize);
  
  for (i = 0; i < Vsize; i++) {
    vec[i] = (double)(v1[i]+v2[i]) / (double)2.0;
  }
  return vec; 
}  /* CalculateVectorAverage() */

/* ------------------------------------------------------------------ */

/* Partial Distortion Search used to speedup distance calculation */
double Distance(double* v1, CODEBOOK* CB, int v2, double maxdist) 
{
  int i;
  double diff, dist = 0.0;

  for (i = 0; (i < VectorSize(CB)) && (dist < maxdist); i++) {
    diff  = v1[i] - VectorScalar(CB,v2,i);
    dist += diff*diff;
  }
  
  return dist;
}  /* Distance() */

/* ------------------------------------------------------------------ */

/* tell if clusters c1 and c2 are neighbour clusters */
int AreNeighbourClusters(CODEBOOK* CB, int c1, int c2) 
{
  int i;
  double d1, d2;
  double *average;
	  
  if (c1 == c2) {
    /* cluster *is not* neighbour of herself */ 
    return 0;
  }

  average = CalculateVectorAverage(Vector(CB,c1),Vector(CB,c2),VectorSize(CB));
  
  d1 = Distance(average,CB,c1,DBL_MAX);
  
  for(i = 0; i < BookSize(CB); i++) {
    if (i != c1 && i != c2) {
      d2 = Distance(average,CB,i,d1);
      
      if (d2 < d1) {
        /* no neighbours, because there is cluster i between them! */
        free(average);
        return 0;
      }
    }
  }

  free(average);
  return 1;
}  /* AreNeighbourClusters() */

/* ------------------------------------------------------------------ */

STACK **FindNeighbours(CODEBOOK* CB)
{
  int    i,j;
  int    *p = NULL;
  STACK  **neighbours = NULL;

  neighbours = (STACK**)allocate(sizeof(STACK*)*BookSize(CB));
  for (i = 0; i < BookSize(CB); i++) {
    neighbours[i] = S_make();
  }

  for (i = 0; i < BookSize(CB); i++) {
    for  (j = 0; j <= i; j++) {
      if (AreNeighbourClusters(CB,i,j)) {
        p  = (int*)allocate(sizeof(int));
	*p = j;
        S_push(neighbours[i], (void*)p);
	
        p  = (int*)allocate(sizeof(int));
	*p = i;
        S_push(neighbours[j], (void*)p);
      }
    }
  }

  return neighbours;
}  /* FindNeighbours() */

/* ------------------------------------------------------------------ */

void FreeNeighbours(CODEBOOK* CB, STACK** neighbours)
{
  int i;

  for (i = 0; i < BookSize(CB); i++) {
    S_free(neighbours[i]);
  }

  free(neighbours);
}  /* FreeNeighbours() */


Graph *VoronoiNeighbGraph(TRAININGSET* TS, int maxNeighbours)
{
  Graph *g;
  GraphVector *v;
  int K;
  STACK** neighbours;
  int*    p = NULL;
  int*    p2 = NULL;
  STACK** neighbours2 = NULL;
  STACK*  S;
  int     i,j;
  int     maxNeighboursInStack, thisNeighbours;

  neighbours = FindNeighbours(TS);

  /* Let's copy the stack to another stack to know the maximum number of 
     neighbours */
  neighbours2 = (STACK**)allocate(sizeof(STACK*)*BookSize(TS));
  for (i = 0; i < BookSize(TS); i++) {
    neighbours2[i] = S_make();
  }

  maxNeighboursInStack = 0;
  for (i = 0; i < BookSize(TS); i++) {
    S = neighbours[i];
    thisNeighbours = 0;
    while (!S_empty(S)) {
      thisNeighbours++;
      p2 = (int*)S_pop(S);
      p  = (int*)allocate(sizeof(int));
      *p = *p2;
      S_push(neighbours2[i], (void*)p);
    }
    if (thisNeighbours>maxNeighboursInStack) 
      maxNeighboursInStack = thisNeighbours; 
  }



  K = maxNeighbours;
  /* There are max BookSize(TS)-1 Voronoi neighbours for a vector */
  if ((BookSize(TS)-1)<maxNeighbours) K = BookSize(TS)-1;
  if (maxNeighboursInStack<K) K = maxNeighboursInStack;  

  g = AllocateMemoryForGraph(BookSize(TS), K, VectorSize(TS));

  for ( i = 0; i < BookSize(TS); i++ ) {
      v = AllocateMemoryForGraphVector(g, i);
      for ( j = 0; j < VectorSize(TS); j++) {
        GraphPutVectorElement(v, j, VectorScalar(TS,i,j));
      }
      S = neighbours2[i];
      for ( j = 0; j < K; j++ ) {
	if (!S_empty(S)) {
	  p = (int*)S_pop(S);
          GraphPutNN(g, v, *p, j);
	}
	else GraphPutNN(g, v, i, j);  /* put the vector as its own neighbor,
                                         this is how we fill up to K-1
                                         neighbours */
      }
      /*printf("%p  %p\n", v, g->vectors[i]);*/
      g->vectors[i] = v;
  }
  return g;
}
