/*-------------------------------------------------------------------*/
/* xNN.C                           Mikko Malinen                    */
/*                                                                   */
/* Console interface for graph forming routines.      		     */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "xNN"
#define VersionNumber  "Version 0.03"
#define LastUpdated    "13.02.17" /* Sami Sieranoja */

/* ----------------------------------------------------------------- */
/* Version history:                                                  */
/* v0.01 05.08.09 Initial version (with name cbnn) (MM)              */
/* v0.02 29.03.14 Many separate options combined (MM)                */
/* v0.03 13.02.17 Compile with new modules dir (SS)                  */
/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <time.h>
#include <values.h>

#define  FACTFILE  "xNN.fac"

/* Comment the next line out if not used */
/*# include "pnnfast.h"*/
#include "pnngraph.h"

#include "parametr.c"
#include "interfc.h"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "random.h"
#include "solution.h"
#include "sortcb.h"
#include "MST_Prim.h"
#include "stack.h"   /* For Voronoi */
#include "neighb.h"  /* Voronoi */
#include "FastMST.h"
#include "kMSTKruskal.h"
#include "MSTKruskal.h"
#include "kmeans.h"  /* needed in FastkNN */
#include "FastkNN.h"

/*--------------------------- Constants -----------------------------*/

#define  MAXFILENAME    64
#define  InitialSolutionGiven   (InitCBName[0] != 0x00)


/* =================  T I M E   F U N C T I O N S  ================= */


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long tout,h,m,s;

  tout = t;
  s = t % 60;
  t = t / 60;
  m = t % 60;
  h = t / 60;

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:  printf(" %li\n",tout);
             break;
    case 2:
    case 3:
    case 4:  printf("  %li:%02li:%02li\n", h, m, s);
             break;
    default: printf("Time: %li:%02li:%02li\n", h, m, s);
             break;
    }
}


/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("xNN - Graph creation program.\n");
  printf("Usage: %s [%coption] <training set> <codebook> <output codebook> <graph file>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s bridge solution1 solution_out graph_out \n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* TSName,
                               char* InitCBName,
                               char* CBName,
                               char* PAName)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
	printf("Training Set:        %s\n", TSName);
    if(InitialSolutionGiven)
    printf("Initial codebook:    %s\n", InitCBName);
    printf("Output codebook:     %s\n", CBName);
    if(Value(SavePartition))
    printf("Output partition:    %s\n", PAName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintResult(long watch, double error)
{
  switch( Value(QuietLevel) )
	 {
     case 0:  break;
     case 1:
     case 2:  printf("%9.4f  ", error); break;
     default: printf("Distortion: %-9.4f\n", error); break;
	 }
  PrintTime(watch);
}


/*======================  SOLUTION HANDLING  ========================*/


static void GenerateInitialSolution(char*	  InitCBName,
                                    TRAININGSET*  TS,
                                    CODEBOOK*     CB,
                                    PARTITIONING* P)
/* Currently initial solution can only be CODEBOOK. It should be
   possible to input also partition as an input solution. */
{
  if( InitialSolutionGiven )
	 {
	 ReadCodebook(InitCBName, CB);
	 CreateNewPartitioning(P, TS, BookSize(CB));
	 GenerateOptimalPartitioning(TS, CB, P);
	 AddGenerationMethod(CB, "PNN from init");
	 }
  else	/* Generate initial codebook from TS */
    {
	CreateNewCodebook(CB, BookSize(TS), TS);
    CopyCodebook(TS, CB);
    SetVersionString(CB, CBFILE);
	CreateNewPartitioning(P, TS, BookSize(CB));
    PutAllInOwnPartition(TS, P);
    AddGenerationMethod(CB, "PNN");
  }
}


/*-------------------------------------------------------------------*/


static void SaveSolution(char*         CBName,
                         char*         PAName,
                         TRAININGSET*  TS,
                         CODEBOOK*     CB,
                         PARTITIONING* P)
{
  int ndups;

  ndups = DuplicatesInCodebook(CB);
  if( ndups>0 && Value(QuietLevel)>=2 )
    {
    printf("WARNING: Final CB contains %i duplicates.\n", ndups);
    }
  SortCodebook(CB, DATA_DESCENDING);
  WriteCodebook(CBName, CB, Value(OverWrite));
  if(Value(SavePartition))
     {
     WritePartitioning(PAName, P, TS, Value(OverWrite));
     }
}


/*-------------------------------------------------------------------*/


/*===========================  M A I N ==============================*/


int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  CODEBOOK      CB;
  PARTITIONING  P;
  Graph*        g = NULL;
  char          TSName[MAXFILENAME]      = { '\0' };
  char          InitCBName[MAXFILENAME]  = { '\0' };
  char          CBName[MAXFILENAME]      = { '\0' };
  char          PAName[MAXFILENAME]      = { '\0' };
  char          DGRName[MAXFILENAME]     = { '\0' };
  double        Error=0.0;
  long          watch;

  ParameterInfo paraminfo[4] = { { TSName,     FormatNameTS, 0, INFILE  },
                                 { InitCBName, FormatNameCB, 1, INFILE  },
			         { CBName,     FormatNameCB, 0, OUTFILE },
				 { DGRName,    "dgr",        0, OUTFILE } };

  ParseParameters(argc, argv, 4, paraminfo);
  PickFileName(CBName, PAName);
  CheckFileName(PAName, FormatNamePA);
  if (Value(GraphType)==NOTHING)
    {
      ErrorMessage("\nERROR: You must select one graph type.\n\n");
      ExitProcessing(FATAL_ERROR);
    }

  PrintOperatingInfo(TSName, InitCBName, CBName, PAName);
  ReadTrainingSet(TSName, &TS);

  SetWatch(&watch);

  GenerateInitialSolution(InitCBName, &TS, &CB, &P);

  switch( Value(GraphType) )
  {
  case NOTHING:

      break;

  case PNN:
    g = PNNGraph2(&TS, &CB, &P, Value(KNN), Value(CodebookSize), Value(Heap), Value(GCMethod),
          Value(QuietLevel), watch, Value(Pds), Value(Mps), Value(Mlf), Value(Lazy), Value(Cf));
    Error = AverageErrorCBFast(&TS, &CB, &P, MSE);
    PrintResult(watch, PrintableError(Error,&CB));
      break;

  case MST:
    g = MST_Prim(&CB, Value(QuietLevel));
    if (g == 0)
      {
	ErrorMessage("\nERROR occurred during MST construction.\n\n");
	ExitProcessing(FATAL_ERROR);
      }
    break;

  case kMST:
    g = kMSTKruskal(&CB, Value(kMSTkValue));
    if (g == 0)
      {
	ErrorMessage("\nERROR occurred during kMST construction.\n\n");
	ExitProcessing(FATAL_ERROR);
      }
    break;

  case xNN:
    g = VoronoiNeighbGraph(&CB, BookSize(&CB)-1);
    if (g == 0)
      {
	ErrorMessage("\nERROR occurred during Voronoi-neighbour graph constru\
ction.\n\n");
	ExitProcessing(FATAL_ERROR);
      }
    break;
  case Fast_kNN:
    g = FastkNN(&TS, &CB, &P, Value(KNN), BookSize(&CB), Value(Heap),\
		Value(GCMethod), Value(QuietLevel), watch, Value(Pds),\
		Value(Mps), Value(Mlf), Value(Lazy), Value(Cf));
    if (g == 0)
      {
	ErrorMessage("\nERROR occurred during Fast kNN graph construction.\n\n");
	ExitProcessing(FATAL_ERROR);
      }
    break;
  default:
      printf("\nYou may have a wrong graph type option !\nAllowed options are 0 - 4!\n\n");
      break;
  }




  GraphWrite(DGRName, g, Value(OverWrite));
  //GraphWrite("pnntulos.dgr", g, Value(OverWrite));
  //SaveSolution(CBName, PAName, &TS, &CB, &P);
  FreeCodebook(&TS);
  FreeCodebook(&CB);
  FreePartitioning(&P);
  checkmemory();
  return( 0 );
}

